@@ -49,7 +49,8 @@ const KeybindingMode = {
     LOGIN_SCREEN:  1 << 4,
     MESSAGE_TRAY:  1 << 5,
     SYSTEM_MODAL:  1 << 6,
-    LOOKING_GLASS: 1 << 7
+    LOOKING_GLASS: 1 << 7,
+    ALL:           ~0,
 };
 
 let componentManager = null;
@@ -160,13 +161,13 @@ function start() {
     xdndHandler = new XdndHandler.XdndHandler();
     ctrlAltTabManager = new CtrlAltTab.CtrlAltTabManager();
     overview = new Overview.Overview();
+    wm = new WindowManager.WindowManager();
     magnifier = new Magnifier.Magnifier();
     if (UnlockDialog.isSupported())
         screenShield = new ScreenShield.ScreenShield();
     else
         screenShield = new ScreenShield.ScreenShieldFallback();
     panel = new Panel.Panel();
-    wm = new WindowManager.WindowManager();
     messageTray = new MessageTray.MessageTray();
     keyboard = new Keyboard.Keyboard();
     notificationDaemon = new NotificationDaemon.NotificationDaemon();
@@ -180,7 +181,7 @@ function start() {
     global.screen.override_workspace_layout(Meta.ScreenCorner.TOPLEFT,
                                             false, -1, 1);
     global.display.connect('overlay-key', Lang.bind(overview, overview.toggle));
-    sessionMode.connect('update', _sessionUpdated);
+    sessionMode.connect('updated', _sessionUpdated);
     _sessionUpdated();
 
     // Provide the bus object for gnome-session to
