@@ -1,9 +1,11 @@
 // -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
 
+const Clutter = imports.gi.Clutter;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
 const GnomeDesktop = imports.gi.GnomeDesktop;
 const Lang = imports.lang;
+const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const St = imports.gi.St;
@@ -21,6 +23,7 @@ try {
 const Main = imports.ui.main;
 const PopupMenu = imports.ui.popupMenu;
 const PanelMenu = imports.ui.panelMenu;
+const SwitcherPopup = imports.ui.switcherPopup;
 const Util = imports.misc.util;
 
 const DESKTOP_INPUT_SOURCES_SCHEMA = 'org.gnome.desktop.input-sources';
@@ -47,6 +50,7 @@ const IBusManager = new Lang.Class({
         this._engines = {};
         this._ready = false;
         this._registerPropertiesId = 0;
+        this._currentEngineName = null;
 
         this._nameWatcherId = Gio.DBus.session.watch_name(IBus.SERVICE_IBUS,
                                                           Gio.BusNameWatcherFlags.NONE,
@@ -66,6 +70,10 @@ const IBusManager = new Lang.Class({
         this._engines = {};
         this._ready = false;
         this._registerPropertiesId = 0;
+        this._currentEngineName = null;
+
+        if (this._readyCallback)
+            this._readyCallback(false);
     },
 
     _onNameAppeared: function() {
@@ -89,12 +97,10 @@ const IBusManager = new Lang.Class({
                 let name = enginesList[i].get_name();
                 this._engines[name] = enginesList[i];
             }
+            this._updateReadiness();
         } else {
             this._clear();
-            return;
         }
-
-        this._updateReadiness();
     },
 
     _initPanelService: function(ibus, result) {
@@ -105,27 +111,31 @@ const IBusManager = new Lang.Class({
             this._candidatePopup.setPanelService(this._panelService);
             // Need to set this to get 'global-engine-changed' emitions
             this._ibus.set_watch_ibus_signal(true);
-            this._ibus.connect('global-engine-changed', Lang.bind(this, this._resetProperties));
+            this._ibus.connect('global-engine-changed', Lang.bind(this, this._engineChanged));
             this._panelService.connect('update-property', Lang.bind(this, this._updateProperty));
-            this._resetProperties();
+            // If an engine is already active we need to get its properties
+            this._ibus.get_global_engine_async(-1, null, Lang.bind(this, function(i, result) {
+                let engine = this._ibus.get_global_engine_async_finish(result);
+                if (!engine)
+                    return;
+                this._engineChanged(this._ibus, engine.get_name());
+            }));
+            this._updateReadiness();
         } else {
             this._clear();
-            return;
         }
-
-        this._updateReadiness();
     },
 
     _updateReadiness: function() {
         this._ready = (Object.keys(this._engines).length > 0 &&
                        this._panelService != null);
 
-        if (this._ready && this._readyCallback)
-            this._readyCallback();
+        if (this._readyCallback)
+            this._readyCallback(this._ready);
     },
 
-    _resetProperties: function() {
-        this.emit('properties-registered', null);
+    _engineChanged: function(bus, engineName) {
+        this._currentEngineName = engineName;
 
         if (this._registerPropertiesId != 0)
             return;
@@ -138,25 +148,18 @@ const IBusManager = new Lang.Class({
                 this._panelService.disconnect(this._registerPropertiesId);
                 this._registerPropertiesId = 0;
 
-                this.emit('properties-registered', props);
+                this.emit('properties-registered', this._currentEngineName, props);
             }));
     },
 
     _updateProperty: function(panel, prop) {
-        this.emit('property-updated', prop);
+        this.emit('property-updated', this._currentEngineName, prop);
     },
 
     activateProperty: function(key, state) {
         this._panelService.property_activate(key, state);
     },
 
-    hasProperties: function(id) {
-        if (id == 'anthy')
-            return true;
-
-        return false;
-    },
-
     getEngineDesc: function(id) {
         if (!IBus || !this._ready)
             return null;
@@ -181,16 +184,127 @@ const LayoutMenuItem = new Lang.Class({
     }
 });
 
+const InputSource = new Lang.Class({
+    Name: 'InputSource',
+
+    _init: function(type, id, displayName, shortName, index) {
+        this.type = type;
+        this.id = id;
+        this.displayName = displayName;
+        this._shortName = shortName;
+        this.index = index;
+
+        this._menuItem = new LayoutMenuItem(this.displayName, this._shortName);
+        this._menuItem.connect('activate', Lang.bind(this, this.activate));
+        this._indicatorLabel = new St.Label({ text: this._shortName });
+
+        this.properties = null;
+    },
+
+    destroy: function() {
+        this._menuItem.destroy();
+        this._indicatorLabel.destroy();
+    },
+
+    get shortName() {
+        return this._shortName;
+    },
+
+    set shortName(v) {
+        this._shortName = v;
+        this._menuItem.indicator.set_text(v);
+        this._indicatorLabel.set_text(v);
+    },
+
+    get menuItem() {
+        return this._menuItem;
+    },
+
+    get indicatorLabel() {
+        return this._indicatorLabel;
+    },
+
+    activate: function() {
+        this.emit('activate');
+    },
+});
+Signals.addSignalMethods(InputSource.prototype);
+
+const InputSourcePopup = new Lang.Class({
+    Name: 'InputSourcePopup',
+    Extends: SwitcherPopup.SwitcherPopup,
+
+    _init: function(items, action, actionBackward) {
+        this.parent(items);
+
+        this._action = action;
+        this._actionBackward = actionBackward;
+    },
+
+    _createSwitcher: function() {
+        this._switcherList = new InputSourceSwitcher(this._items);
+        return true;
+    },
+
+    _initialSelection: function(backward, binding) {
+        if (binding == 'switch-input-source') {
+            if (backward)
+                this._selectedIndex = this._items.length - 1;
+        } else if (binding == 'switch-input-source-backward') {
+            if (!backward)
+                this._selectedIndex = this._items.length - 1;
+        }
+        this._select(this._selectedIndex);
+    },
+
+    _keyPressHandler: function(keysym, backwards, action) {
+        if (action == this._action)
+            this._select(backwards ? this._previous() : this._next());
+        else if (action == this._actionBackward)
+            this._select(backwards ? this._next() : this._previous());
+        else if (keysym == Clutter.Left)
+            this._select(this._previous());
+        else if (keysym == Clutter.Right)
+            this._select(this._next());
+    },
+
+    _finish : function() {
+        this.parent();
+
+        this._items[this._selectedIndex].activate();
+    },
+});
+
+const InputSourceSwitcher = new Lang.Class({
+    Name: 'InputSourceSwitcher',
+    Extends: SwitcherPopup.SwitcherList,
+
+    _init: function(items) {
+        this.parent(true);
+
+        for (let i = 0; i < items.length; i++)
+            this._addIcon(items[i]);
+    },
+
+    _addIcon: function(item) {
+        let box = new St.BoxLayout({ vertical: true });
+
+        let bin = new St.Bin({ style_class: 'input-source-switcher-symbol' });
+        let symbol = new St.Label({ text: item.shortName });
+        bin.set_child(symbol);
+        box.add(bin, { x_fill: false, y_fill: false } );
+
+        let text = new St.Label({ text: item.displayName });
+        box.add(text, { x_fill: false });
+
+        this.addItem(box, text);
+    }
+});
+
 const InputSourceIndicator = new Lang.Class({
     Name: 'InputSourceIndicator',
     Extends: PanelMenu.Button,
 
-    _propertiesWhitelist: [
-        'InputMode',
-        'TypingMode',
-        'DictMode'
-    ],
-
     _init: function() {
         this.parent(0.0, _("Keyboard"));
 
@@ -201,14 +315,36 @@ const InputSourceIndicator = new Lang.Class({
         this.actor.add_actor(this._container);
         this.actor.add_style_class_name('panel-status-button');
 
-        this._labelActors = {};
-        this._layoutItems = {};
+        // All valid input sources currently in the gsettings
+        // KEY_INPUT_SOURCES list indexed by their index there
+        this._inputSources = {};
+        // All valid input sources currently in the gsettings
+        // KEY_INPUT_SOURCES list of type INPUT_SOURCE_TYPE_IBUS
+        // indexed by the IBus ID
+        this._ibusSources = {};
 
+        this._currentSource = null;
+
+        // All valid input sources currently in the gsettings
+        // KEY_INPUT_SOURCES list ordered by most recently used
+        this._mruSources = [];
+        this._keybindingAction =
+            Main.wm.addKeybinding('switch-input-source',
+                                  new Gio.Settings({ schema: "org.gnome.desktop.wm.keybindings" }),
+                                  Meta.KeyBindingFlags.REVERSES,
+                                  Main.KeybindingMode.ALL,
+                                  Lang.bind(this, this._switchInputSource));
+        this._keybindingActionBackward =
+            Main.wm.addKeybinding('switch-input-source-backward',
+                                  new Gio.Settings({ schema: "org.gnome.desktop.wm.keybindings" }),
+                                  Meta.KeyBindingFlags.REVERSES |
+                                  Meta.KeyBindingFlags.REVERSED,
+                                  Main.KeybindingMode.ALL,
+                                  Lang.bind(this, this._switchInputSource));
         this._settings = new Gio.Settings({ schema: DESKTOP_INPUT_SOURCES_SCHEMA });
         this._settings.connect('changed::' + KEY_CURRENT_INPUT_SOURCE, Lang.bind(this, this._currentInputSourceChanged));
         this._settings.connect('changed::' + KEY_INPUT_SOURCES, Lang.bind(this, this._inputSourcesChanged));
 
-        this._currentSourceIndex = this._settings.get_uint(KEY_CURRENT_INPUT_SOURCE);
         this._xkbInfo = new GnomeDesktop.XkbInfo();
 
         this._propSeparator = new PopupMenu.PopupSeparatorMenuItem();
@@ -217,9 +353,8 @@ const InputSourceIndicator = new Lang.Class({
         this.menu.addMenuItem(this._propSection);
         this._propSection.actor.hide();
 
-        this._properties = null;
-
-        this._ibusManager = new IBusManager(Lang.bind(this, this._inputSourcesChanged));
+        this._ibusReady = false;
+        this._ibusManager = new IBusManager(Lang.bind(this, this._ibusReadyCallback));
         this._ibusManager.connect('properties-registered', Lang.bind(this, this._ibusPropertiesRegistered));
         this._ibusManager.connect('property-updated', Lang.bind(this, this._ibusPropertyUpdated));
         this._inputSourcesChanged();
@@ -241,18 +376,29 @@ const InputSourceIndicator = new Lang.Class({
         this._showLayoutItem.actor.visible = Main.sessionMode.allowSettings;
     },
 
+    _ibusReadyCallback: function(ready) {
+        if (this._ibusReady == ready)
+            return;
+
+        this._ibusReady = ready;
+        this._mruSources = [];
+        this._inputSourcesChanged();
+    },
+
+    _switchInputSource: function(display, screen, window, binding) {
+        let popup = new InputSourcePopup(this._mruSources, this._keybindingAction, this._keybindingActionBackward);
+        let modifiers = binding.get_modifiers();
+        let backwards = modifiers & Meta.VirtualModifier.SHIFT_MASK;
+        if (!popup.show(backwards, binding.get_name(), binding.get_mask()))
+            popup.destroy();
+    },
+
     _currentInputSourceChanged: function() {
-        let nVisibleSources = Object.keys(this._layoutItems).length;
-        let newCurrentSourceIndex = this._settings.get_uint(KEY_CURRENT_INPUT_SOURCE);
-        let newLayoutItem = this._layoutItems[newCurrentSourceIndex];
-        let hasProperties;
+        let nVisibleSources = Object.keys(this._inputSources).length;
+        let newSourceIndex = this._settings.get_uint(KEY_CURRENT_INPUT_SOURCE);
+        let newSource = this._inputSources[newSourceIndex];
 
-        if (newLayoutItem)
-            hasProperties = this._ibusManager.hasProperties(newLayoutItem.ibusEngineId);
-        else
-            hasProperties = false;
-
-        if (!newLayoutItem || (nVisibleSources < 2 && !hasProperties)) {
+        if (!newSource || (nVisibleSources < 2 && !newSource.properties)) {
             // This source index might be invalid if we weren't able
             // to build a menu item for it, so we hide ourselves since
             // we can't fix it here. *shrug*
@@ -266,106 +412,121 @@ const InputSourceIndicator = new Lang.Class({
 
         this.actor.show();
 
-        if (this._layoutItems[this._currentSourceIndex]) {
-            this._layoutItems[this._currentSourceIndex].setShowDot(false);
-            this._container.set_skip_paint(this._labelActors[this._currentSourceIndex], true);
+        let oldSource;
+        [oldSource, this._currentSource] = [this._currentSource, newSource];
+
+        if (oldSource) {
+            oldSource.menuItem.setShowDot(false);
+            this._container.set_skip_paint(oldSource.indicatorLabel, true);
         }
 
-        newLayoutItem.setShowDot(true);
+        newSource.menuItem.setShowDot(true);
+        this._container.set_skip_paint(newSource.indicatorLabel, false);
 
-        let newLabelActor = this._labelActors[newCurrentSourceIndex];
-        this._container.set_skip_paint(newLabelActor, false);
+        this._buildPropSection(newSource.properties);
 
-        if (hasProperties)
-            newLabelActor.set_text(newLayoutItem.indicator.get_text());
-
-        this._currentSourceIndex = newCurrentSourceIndex;
+        for (let i = 1; i < this._mruSources.length; ++i)
+            if (this._mruSources[i] == newSource) {
+                let currentSource = this._mruSources.splice(i, 1);
+                this._mruSources = currentSource.concat(this._mruSources);
+                break;
+            }
     },
 
     _inputSourcesChanged: function() {
         let sources = this._settings.get_value(KEY_INPUT_SOURCES);
         let nSources = sources.n_children();
 
-        for (let i in this._layoutItems)
-            this._layoutItems[i].destroy();
+        for (let i in this._inputSources)
+            this._inputSources[i].destroy();
 
-        for (let i in this._labelActors)
-            this._labelActors[i].destroy();
+        this._inputSources = {};
+        this._ibusSources = {};
 
-        this._layoutItems = {};
-        this._labelActors = {};
-
-        let infos = [];
-        let infosByShortName = {};
+        let inputSourcesByShortName = {};
 
         for (let i = 0; i < nSources; i++) {
-            let info = { exists: false };
+            let displayName;
+            let shortName;
             let [type, id] = sources.get_child_value(i).deep_unpack();
+            let exists = false;
 
             if (type == INPUT_SOURCE_TYPE_XKB) {
-                [info.exists, info.displayName, info.shortName, , ] =
+                [exists, displayName, shortName, , ] =
                     this._xkbInfo.get_layout_info(id);
             } else if (type == INPUT_SOURCE_TYPE_IBUS) {
                 let engineDesc = this._ibusManager.getEngineDesc(id);
                 if (engineDesc) {
                     let language = IBus.get_language_name(engineDesc.get_language());
-
-                    info.exists = true;
-                    info.displayName = language + ' (' + engineDesc.get_longname() + ')';
-                    info.shortName = this._makeEngineShortName(engineDesc);
-                    info.ibusEngineId = id;
+                    exists = true;
+                    displayName = language + ' (' + engineDesc.get_longname() + ')';
+                    shortName = this._makeEngineShortName(engineDesc);
                 }
             }
 
-            if (!info.exists)
+            if (!exists)
                 continue;
 
-            info.sourceIndex = i;
+            let is = new InputSource(type, id, displayName, shortName, i);
 
-            if (!(info.shortName in infosByShortName))
-                infosByShortName[info.shortName] = [];
-            infosByShortName[info.shortName].push(info);
-            infos.push(info);
-        }
-
-        for (let i = 0; i < infos.length; i++) {
-            let info = infos[i];
-            if (infosByShortName[info.shortName].length > 1) {
-                let sub = infosByShortName[info.shortName].indexOf(info) + 1;
-                info.shortName += String.fromCharCode(0x2080 + sub);
-            }
-
-            let item = new LayoutMenuItem(info.displayName, info.shortName);
-            item.ibusEngineId = info.ibusEngineId;
-            this._layoutItems[info.sourceIndex] = item;
-            this.menu.addMenuItem(item, i);
-            item.connect('activate', Lang.bind(this, function() {
+            is.connect('activate', Lang.bind(this, function() {
                 this._settings.set_value(KEY_CURRENT_INPUT_SOURCE,
-                                         GLib.Variant.new_uint32(info.sourceIndex));
+                                         GLib.Variant.new_uint32(is.index));
             }));
 
-            let shortLabel = new St.Label({ text: info.shortName });
-            this._labelActors[info.sourceIndex] = shortLabel;
-            this._container.add_actor(shortLabel);
-            this._container.set_skip_paint(shortLabel, true);
+            if (!(is.shortName in inputSourcesByShortName))
+                inputSourcesByShortName[is.shortName] = [];
+            inputSourcesByShortName[is.shortName].push(is);
+
+            this._inputSources[is.index] = is;
+
+            if (is.type == INPUT_SOURCE_TYPE_IBUS)
+                this._ibusSources[is.id] = is;
         }
 
+        let menuIndex = 0;
+        for (let i in this._inputSources) {
+            let is = this._inputSources[i];
+            if (inputSourcesByShortName[is.shortName].length > 1) {
+                let sub = inputSourcesByShortName[is.shortName].indexOf(is) + 1;
+                is.shortName += String.fromCharCode(0x2080 + sub);
+            }
+
+            this.menu.addMenuItem(is.menuItem, menuIndex++);
+
+            this._container.add_actor(is.indicatorLabel);
+            this._container.set_skip_paint(is.indicatorLabel, true);
+        }
+
+        let sourcesList = [];
+        for (let i in this._inputSources)
+            sourcesList.push(this._inputSources[i]);
+
+        let mruSources = [];
+        for (let i = 0; i < this._mruSources.length; i++) {
+            for (let j = 0; j < sourcesList.length; j++)
+                if (this._mruSources[i].type == sourcesList[j].type &&
+                    this._mruSources[i].id == sourcesList[j].id) {
+                    mruSources = mruSources.concat(sourcesList.splice(j, 1));
+                    break;
+                }
+        }
+        this._mruSources = mruSources.concat(sourcesList);
+
         this._currentInputSourceChanged();
     },
 
     _showLayout: function() {
         Main.overview.hide();
 
-        let sources = this._settings.get_value(KEY_INPUT_SOURCES);
-        let current = this._settings.get_uint(KEY_CURRENT_INPUT_SOURCE);
-        let [type, id] = sources.get_child_value(current).deep_unpack();
+        let source = this._currentSource;
         let xkbLayout = '';
         let xkbVariant = '';
 
-        if (type == INPUT_SOURCE_TYPE_XKB) {
-            [, , , xkbLayout, xkbVariant] = this._xkbInfo.get_layout_info(id);
-        } else if (type == INPUT_SOURCE_TYPE_IBUS) {
-            let engineDesc = this._ibusManager.getEngineDesc(id);
+        if (source.type == INPUT_SOURCE_TYPE_XKB) {
+            [, , , xkbLayout, xkbVariant] = this._xkbInfo.get_layout_info(source.id);
+        } else if (source.type == INPUT_SOURCE_TYPE_IBUS) {
+            let engineDesc = this._ibusManager.getEngineDesc(source.id);
             if (engineDesc) {
                 xkbLayout = engineDesc.get_layout();
                 xkbVariant = '';
@@ -394,26 +555,25 @@ const InputSourceIndicator = new Lang.Class({
         return String.fromCharCode(0x2328); // keyboard glyph
     },
 
-    _propertyWhitelisted: function(prop) {
-        for (let i = 0; i < this._propertiesWhitelist.length; ++i) {
-            let key = prop.get_key();
-            if (key.substr(0, this._propertiesWhitelist[i].length) == this._propertiesWhitelist[i])
-                return true;
-        }
-        return false;
-    },
-
-    _ibusPropertiesRegistered: function(im, props) {
-        this._properties = props;
-        this._buildPropSection();
-    },
-
-    _ibusPropertyUpdated: function(im, prop) {
-        if (!this._propertyWhitelisted(prop))
+    _ibusPropertiesRegistered: function(im, engineName, props) {
+        let source = this._ibusSources[engineName];
+        if (!source)
             return;
 
-        if (this._updateSubProperty(this._properties, prop))
-            this._buildPropSection();
+        source.properties = props;
+
+        if (source == this._currentSource)
+            this._currentInputSourceChanged();
+    },
+
+    _ibusPropertyUpdated: function(im, engineName, prop) {
+        let source = this._ibusSources[engineName];
+        if (!source)
+            return;
+
+        if (this._updateSubProperty(source.properties, prop) &&
+            source == this._currentSource)
+            this._currentInputSourceChanged();
     },
 
     _updateSubProperty: function(props, prop) {
@@ -433,25 +593,12 @@ const InputSourceIndicator = new Lang.Class({
         return false;
     },
 
-    _updateIndicatorLabel: function(text) {
-        let layoutItem = this._layoutItems[this._currentSourceIndex];
-        let hasProperties;
-
-        if (layoutItem)
-            hasProperties = this._ibusManager.hasProperties(layoutItem.ibusEngineId);
-        else
-            hasProperties = false;
-
-        if (hasProperties)
-            this._labelActors[this._currentSourceIndex].set_text(text);
-    },
-
-    _buildPropSection: function() {
+    _buildPropSection: function(properties) {
         this._propSeparator.actor.hide();
         this._propSection.actor.hide();
         this._propSection.removeAll();
 
-        this._buildPropSubMenu(this._propSection, this._properties);
+        this._buildPropSubMenu(this._propSection, properties);
 
         if (!this._propSection.isEmpty()) {
             this._propSection.actor.show();
@@ -468,8 +615,7 @@ const InputSourceIndicator = new Lang.Class({
         for (let i = 0; (p = props.get(i)) != null; ++i) {
             let prop = p;
 
-            if (!this._propertyWhitelisted(prop) ||
-                !prop.get_visible())
+            if (!prop.get_visible())
                 continue;
 
             if (prop.get_key() == 'InputMode') {
@@ -480,15 +626,17 @@ const InputSourceIndicator = new Lang.Class({
                     text = prop.get_label().get_text();
 
                 if (text && text.length > 0 && text.length < 3)
-                    this._updateIndicatorLabel(text);
+                    this._currentSource.indicatorLabel.set_text(text);
             }
 
             let item;
-            let type = prop.get_prop_type();
-            if (type == IBus.PropType.MENU) {
+            switch (prop.get_prop_type()) {
+            case IBus.PropType.MENU:
                 item = new PopupMenu.PopupSubMenuMenuItem(prop.get_label().get_text());
                 this._buildPropSubMenu(item.menu, prop.get_sub_props());
-            } else if (type == IBus.PropType.RADIO) {
+                break;
+
+            case IBus.PropType.RADIO:
                 item = new PopupMenu.PopupMenuItem(prop.get_label().get_text());
                 item.prop = prop;
                 radioGroup.push(item);
@@ -513,7 +661,39 @@ const InputSourceIndicator = new Lang.Class({
                         }
                     }
                 }));
-            } else {
+                break;
+
+            case IBus.PropType.TOGGLE:
+                item = new PopupMenu.PopupSwitchMenuItem(prop.get_label().get_text(), prop.get_state() == IBus.PropState.CHECKED);
+                item.prop = prop;
+                item.connect('toggled', Lang.bind(this, function() {
+                    if (item.state) {
+                        item.prop.set_state(IBus.PropState.CHECKED);
+                        this._ibusManager.activateProperty(item.prop.get_key(),
+                                                           IBus.PropState.CHECKED);
+                    } else {
+                        item.prop.set_state(IBus.PropState.UNCHECKED);
+                        this._ibusManager.activateProperty(item.prop.get_key(),
+                                                           IBus.PropState.UNCHECKED);
+                    }
+                }));
+                break;
+
+            case IBus.PropType.NORMAL:
+                item = new PopupMenu.PopupMenuItem(prop.get_label().get_text());
+                item.prop = prop;
+                item.connect('activate', Lang.bind(this, function() {
+                    this._ibusManager.activateProperty(item.prop.get_key(),
+                                                       IBus.PropState.CHECKED);
+                }));
+                break;
+
+            case IBus.PropType.SEPARATOR:
+                item = new PopupMenu.PopupSeparatorMenuItem();
+                break;
+
+            default:
+                log ('IBus property %s has invalid type %d'.format(prop.get_key(), type));
                 continue;
             }
 
@@ -528,8 +708,9 @@ const InputSourceIndicator = new Lang.Class({
         // for those we don't actually display.
         let max_min_width = 0, max_natural_width = 0;
 
-        for (let i in this._labelActors) {
-            let [min_width, natural_width] = this._labelActors[i].get_preferred_width(for_height);
+        for (let i in this._inputSources) {
+            let is = this._inputSources[i];
+            let [min_width, natural_width] = is.indicatorLabel.get_preferred_width(for_height);
             max_min_width = Math.max(max_min_width, min_width);
             max_natural_width = Math.max(max_natural_width, natural_width);
         }
@@ -541,8 +722,9 @@ const InputSourceIndicator = new Lang.Class({
     _containerGetPreferredHeight: function(container, for_width, alloc) {
         let max_min_height = 0, max_natural_height = 0;
 
-        for (let i in this._labelActors) {
-            let [min_height, natural_height] = this._labelActors[i].get_preferred_height(for_width);
+        for (let i in this._inputSources) {
+            let is = this._inputSources[i];
+            let [min_height, natural_height] = is.indicatorLabel.get_preferred_height(for_width);
             max_min_height = Math.max(max_min_height, min_height);
             max_natural_height = Math.max(max_natural_height, natural_height);
         }
@@ -558,7 +740,9 @@ const InputSourceIndicator = new Lang.Class({
         box.y2 -= box.y1;
         box.y1 = 0;
 
-        for (let i in this._labelActors)
-            this._labelActors[i].allocate_align_fill(box, 0.5, 0, false, false, flags);
+        for (let i in this._inputSources) {
+            let is = this._inputSources[i];
+            is.indicatorLabel.allocate_align_fill(box, 0.5, 0, false, false, flags);
+        }
     }
 });
