@@ -5,6 +5,7 @@ const Gio = imports.gi.Gio;
 const Lang = imports.lang;
 const NetworkManager = imports.gi.NetworkManager;
 const NMClient = imports.gi.NMClient;
+const NMGtk = imports.gi.NMGtk;
 const Signals = imports.signals;
 const St = imports.gi.St;
 
@@ -139,46 +140,6 @@ const NMNetworkMenuItem = new Lang.Class({
     }
 });
 
-const NMWiredSectionTitleMenuItem = new Lang.Class({
-    Name: 'NMWiredSectionTitleMenuItem',
-    Extends: PopupMenu.PopupSwitchMenuItem,
-
-    _init: function(label, params) {
-        params = params || { };
-        params.style_class = 'popup-subtitle-menu-item';
-        this.parent(label, false, params);
-    },
-
-    updateForDevice: function(device) {
-        if (device) {
-            this._device = device;
-            this.setStatus(device.getStatusLabel());
-            this.setToggleState(device.connected);
-        } else
-            this.setStatus('');
-    },
-
-    activate: function(event) {
-        this.parent(event);
-
-        if (!this._device) {
-            log('Section title activated when there is more than one device, should be non reactive');
-            return;
-        }
-
-        let newState = this._switch.state;
-
-        let ok;
-        if (newState)
-            ok = this._device.activate();
-        else
-            ok = this._device.deactivate();
-
-        if (!ok)
-            this._switch.setToggleState(false);
-    }
-});
-
 const NMWirelessSectionTitleMenuItem = new Lang.Class({
     Name: 'NMWirelessSectionTitleMenuItem',
     Extends: PopupMenu.PopupSwitchMenuItem,
@@ -351,7 +312,7 @@ const NMDevice = new Lang.Class({
         this._autoConnectionItem = null;
         this._overflowItem = null;
 
-        this.statusItem = new PopupMenu.PopupSwitchMenuItem(this._getDescription(), this.connected, { style_class: 'popup-subtitle-menu-item' });
+        this.statusItem = new PopupMenu.PopupSwitchMenuItem('', this.connected, { style_class: 'popup-subtitle-menu-item' });
         this._statusChanged = this.statusItem.connect('toggled', Lang.bind(this, function(item, state) {
             let ok;
             if (state)
@@ -512,6 +473,10 @@ const NMDevice = new Lang.Class({
         }
     },
 
+    syncDescription: function() {
+        this.statusItem.label.text = this.device._description;
+    },
+
     // protected
     _createAutomaticConnection: function() {
         throw new TypeError('Invoking pure virtual function NMDevice.createAutomaticConnection');
@@ -609,7 +574,11 @@ const NMDevice = new Lang.Class({
             this.emit('network-lost');
         }
 
-        if (newstate == NetworkManager.DeviceState.FAILED) {
+        /* Emit a notification if activation fails, but don't do it
+           if the reason is no secrets, as that indicates the user
+           cancelled the agent dialog */
+        if (newstate == NetworkManager.DeviceState.FAILED &&
+            reason != NetworkManager.DeviceStateReason.NO_SECRETS) {
             this.emit('activation-failed', reason);
         }
 
@@ -639,34 +608,14 @@ const NMDevice = new Lang.Class({
         this.statusItem.setStatus(this.getStatusLabel());
 
         this.emit('state-changed');
-    },
-
-    _getDescription: function() {
-        let dev_product = this.device.get_product();
-        let dev_vendor = this.device.get_vendor();
-        if (!dev_product || !dev_vendor)
-	    return '';
-
-        let product = Util.fixupPCIDescription(dev_product);
-        let vendor = Util.fixupPCIDescription(dev_vendor);
-        let out = '';
-
-        // Another quick hack; if all of the fixed up vendor string
-        // is found in product, ignore the vendor.
-        if (product.indexOf(vendor) == -1)
-            out += vendor + ' ';
-        out += product;
-
-        return out;
     }
 });
 
-const NMDeviceWired = new Lang.Class({
-    Name: 'NMDeviceWired',
+const NMDeviceSimple = new Lang.Class({
+    Name: 'NMDeviceSimple',
     Extends: NMDevice,
 
     _init: function(client, device, connections) {
-        this._autoConnectionName = _("Auto Ethernet");
         this.category = NMConnectionCategory.WIRED;
 
         this.parent(client, device, connections);
@@ -681,6 +630,18 @@ const NMDeviceWired = new Lang.Class({
         // we can do it here because addConnection and removeConnection
         // both call _createSection at some point
         this.section.actor.visible = this._connections.length > 1;
+    }
+});
+
+const NMDeviceWired = new Lang.Class({
+    Name: 'NMDeviceWired',
+    Extends: NMDeviceSimple,
+
+    _init: function(client, device, connections) {
+        this._autoConnectionName = _("Auto Ethernet");
+        this.category = NMConnectionCategory.WIRED;
+
+        this.parent(client, device, connections);
     },
 
     _createAutomaticConnection: function() {
@@ -867,10 +828,6 @@ const NMDeviceBluetooth = new Lang.Class({
         this._clearSection();
         this._queueCreateSection();
         this._updateStatusItem();
-    },
-
-    _getDescription: function() {
-        return this.device.name || _("Bluetooth");
     }
 });
 
@@ -1598,7 +1555,8 @@ const NMVPNSection = new Lang.Class({
     },
 
     _connectionStateChanged: function(vpnConnection, newstate, reason) {
-        if (newstate == NetworkManager.VPNConnectionState.FAILED) {
+        if (newstate == NetworkManager.VPNConnectionState.FAILED &&
+            reason != NetworkManager.VPNConnectionStateReason.NO_SECRETS) {
             // FIXME: if we ever want to show something based on reason,
             // we need to convert from NetworkManager.VPNConnectionStateReason
             // to NetworkManager.DeviceStateReason
@@ -1651,15 +1609,14 @@ const NMApplet = new Lang.Class({
         this._mobileUpdateId = 0;
         this._mobileUpdateDevice = null;
 
+        this._nmDevices = [];
         this._devices = { };
 
         this._devices.wired = {
             section: new PopupMenu.PopupMenuSection(),
             devices: [ ],
-            item: new NMWiredSectionTitleMenuItem(_("Wired"))
         };
 
-        this._devices.wired.section.addMenuItem(this._devices.wired.item);
         this._devices.wired.section.actor.hide();
         this.menu.addMenuItem(this._devices.wired.section);
         this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
@@ -1667,7 +1624,7 @@ const NMApplet = new Lang.Class({
         this._devices.wireless = {
             section: new PopupMenu.PopupMenuSection(),
             devices: [ ],
-            item: this._makeToggleItem('wireless', _("Wireless"))
+            item: this._makeToggleItem('wireless', _("Wi-Fi"))
         };
         this._devices.wireless.section.addMenuItem(this._devices.wireless.item);
         this._devices.wireless.section.actor.hide();
@@ -1677,9 +1634,7 @@ const NMApplet = new Lang.Class({
         this._devices.wwan = {
             section: new PopupMenu.PopupMenuSection(),
             devices: [ ],
-            item: this._makeToggleItem('wwan', _("Mobile broadband"))
         };
-        this._devices.wwan.section.addMenuItem(this._devices.wwan.item);
         this._devices.wwan.section.actor.hide();
         this.menu.addMenuItem(this._devices.wwan.section);
         this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
@@ -1696,6 +1651,7 @@ const NMApplet = new Lang.Class({
         this._dtypes[NetworkManager.DeviceType.WIFI] = NMDeviceWireless;
         this._dtypes[NetworkManager.DeviceType.MODEM] = NMDeviceModem;
         this._dtypes[NetworkManager.DeviceType.BT] = NMDeviceBluetooth;
+        this._dtypes[NetworkManager.DeviceType.INFINIBAND] = NMDeviceSimple;
         // TODO: WiMax support
 
         // Connection types
@@ -1707,6 +1663,7 @@ const NMApplet = new Lang.Class({
         this._ctypes[NetworkManager.SETTING_BLUETOOTH_SETTING_NAME] = NMConnectionCategory.WWAN;
         this._ctypes[NetworkManager.SETTING_CDMA_SETTING_NAME] = NMConnectionCategory.WWAN;
         this._ctypes[NetworkManager.SETTING_GSM_SETTING_NAME] = NMConnectionCategory.WWAN;
+        this._ctypes[NetworkManager.SETTING_INFINIBAND_SETTING_NAME] = NMConnectionCategory.WIRED;
         this._ctypes[NetworkManager.SETTING_VPN_SETTING_NAME] = NMConnectionCategory.VPN;
 
         this._settings = NMClient.RemoteSettings.new(null);
@@ -1762,16 +1719,23 @@ const NMApplet = new Lang.Class({
             section.actor.hide();
         else {
             section.actor.show();
-            if (devices.length == 1) {
-                let dev = devices[0];
-                dev.statusItem.actor.hide();
-                item.updateForDevice(dev);
-            } else {
-                devices.forEach(function(dev) {
-                    dev.statusItem.actor.show();
-                });
-                // remove status text from the section title item
-                item.updateForDevice(null);
+
+            // Sync the relation between the section title
+            // item (the one with the airplane mode switch)
+            // and the individual device switches
+            if (item) {
+                if (devices.length == 1) {
+                    let dev = devices[0];
+                    dev.statusItem.actor.hide();
+                    item.updateForDevice(dev);
+                } else {
+                    devices.forEach(function(dev) {
+                        dev.statusItem.actor.show();
+                    });
+
+                    // remove status text from the section title item
+                    item.updateForDevice(null);
+                }
             }
         }
     },
@@ -1779,8 +1743,9 @@ const NMApplet = new Lang.Class({
     _readDevices: function() {
         let devices = this._client.get_devices() || [ ];
         for (let i = 0; i < devices.length; ++i) {
-            this._deviceAdded(this._client, devices[i]);
+            this._deviceAdded(this._client, devices[i], true);
         }
+        this._syncDeviceNames();
     },
 
     _notifyForDevice: function(device, iconName, title, text, urgency) {
@@ -1828,7 +1793,18 @@ const NMApplet = new Lang.Class({
         return wrapper;
     },
 
-    _deviceAdded: function(client, device) {
+    _syncDeviceNames: function() {
+        let names = NMGtk.utils_disambiguate_device_names(this._nmDevices);
+        for (let i = 0; i < this._nmDevices.length; i++) {
+            let device = this._nmDevices[i];
+            if (device._description != names[i]) {
+                device._description = names[i];
+                device._delegate.syncDescription();
+            }
+        }
+    },
+
+    _deviceAdded: function(client, device, skipSyncDeviceNames) {
         if (device._delegate) {
             // already seen, not adding again
             return;
@@ -1839,10 +1815,14 @@ const NMApplet = new Lang.Class({
             let section = this._devices[wrapper.category].section;
             let devices = this._devices[wrapper.category].devices;
 
-            section.addMenuItem(wrapper.section, 1);
-            section.addMenuItem(wrapper.statusItem, 1);
+            section.addMenuItem(wrapper.statusItem);
+            section.addMenuItem(wrapper.section);
             devices.push(wrapper);
 
+            this._nmDevices.push(device);
+            if (!skipSyncDeviceNames)
+                this._syncDeviceNames();
+
             this._syncSectionTitle(wrapper.category);
         }
     },
@@ -1860,6 +1840,10 @@ const NMApplet = new Lang.Class({
         let pos = devices.indexOf(wrapper);
         devices.splice(pos, 1);
 
+        pos = this._nmDevices.indexOf(device);
+        this._nmDevices.splice(pos, 1);
+        this._syncDeviceNames();
+
         this._syncSectionTitle(wrapper.category)
     },
 
